# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------

from msrest.serialization import Model


class Quota(Model):
    """Quota information.

    :param dimension: The name of the dimension associated with the quota.
    :type dimension: str
    :param scope: The scope at which the quota is applied. Possible values
     include: 'Workspace', 'Subscription'
    :type scope: str or ~azure.quantum.models.DimensionScope
    :param provider_id: The unique identifier for the provider.
    :type provider_id: str
    :param utilization: The amount of the usage that has been applied for the
     current period.
    :type utilization: float
    :param holds: The amount of the usage that has been reserved but not
     applied for the current period.
    :type holds: float
    :param limit: The maximum amount of usage allowed for the current period.
    :type limit: float
    :param period: The time period in which the quota's underlying meter is
     accumulated. Based on calendar year. 'None' is used for concurrent quotas.
     Possible values include: 'None', 'Monthly'
    :type period: str or ~azure.quantum.models.MeterPeriod
    """

    _attribute_map = {
        'dimension': {'key': 'dimension', 'type': 'str'},
        'scope': {'key': 'scope', 'type': 'str'},
        'provider_id': {'key': 'providerId', 'type': 'str'},
        'utilization': {'key': 'utilization', 'type': 'float'},
        'holds': {'key': 'holds', 'type': 'float'},
        'limit': {'key': 'limit', 'type': 'float'},
        'period': {'key': 'period', 'type': 'str'},
    }

    def __init__(self, **kwargs):
        super(Quota, self).__init__(**kwargs)
        self.dimension = kwargs.get('dimension', None)
        self.scope = kwargs.get('scope', None)
        self.provider_id = kwargs.get('provider_id', None)
        self.utilization = kwargs.get('utilization', None)
        self.holds = kwargs.get('holds', None)
        self.limit = kwargs.get('limit', None)
        self.period = kwargs.get('period', None)



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

from typing import List, Union, Optional

__all__ = ['Term']

class Term:

    def __init__(self, indices: List[int] = None, w: Optional[Union[int, float]] = None, c: Optional[Union[int, float]] = None):
        if(type(w) == None and type(c) == None):
            raise RuntimeError("Cost should be provided for each term.")

        if(w != None):
            if(type(w) != int and type(w) != float):
                raise RuntimeError("w must be a float or int value.")
            else:
                self.c = w
        elif(c != None):
            if(type(c) != int and type(c) != float):
                raise RuntimeError("c must be a float or int value.")
            else:
                self.c = c

        self.ids = indices

    def to_dict(self):
        return self.__dict__

    def __repr__(self):
        return str(self.__dict__)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------

import uuid
from msrest.pipeline import ClientRawResponse

from .. import models


class StorageOperations(object):
    """StorageOperations operations.

    :param client: Client for service requests.
    :param config: Configuration of service client.
    :param serializer: An object model serializer.
    :param deserializer: An object model deserializer.
    """

    models = models

    def __init__(self, client, config, serializer, deserializer):

        self._client = client
        self._serialize = serializer
        self._deserialize = deserializer

        self.config = config

    def sas_uri(
            self, blob_details, custom_headers=None, raw=False, **operation_config):
        """Gets a URL with SAS token for a container/blob in the storage account
        associated with the workspace. The SAS URL can be used to upload job
        input and/or download job output.

        :param blob_details: The details (name and container) of the blob to
         store or download data.
        :type blob_details: ~azure.quantum.models.BlobDetails
        :param dict custom_headers: headers that will be added to the request
        :param bool raw: returns the direct response alongside the
         deserialized response
        :param operation_config: :ref:`Operation configuration
         overrides<msrest:optionsforoperations>`.
        :return: SasUriResponse or ClientRawResponse if raw=true
        :rtype: ~azure.quantum.models.SasUriResponse or
         ~msrest.pipeline.ClientRawResponse
        :raises:
         :class:`RestErrorException<azure.quantum.models.RestErrorException>`
        """
        # Construct URL
        url = self.sas_uri.metadata['url']
        path_format_arguments = {
            'subscriptionId': self._serialize.url("self.config.subscription_id", self.config.subscription_id, 'str'),
            'resourceGroupName': self._serialize.url("self.config.resource_group_name", self.config.resource_group_name, 'str'),
            'workspaceName': self._serialize.url("self.config.workspace_name", self.config.workspace_name, 'str')
        }
        url = self._client.format_url(url, **path_format_arguments)

        # Construct parameters
        query_parameters = {}

        # Construct headers
        header_parameters = {}
        header_parameters['Accept'] = 'application/json'
        header_parameters['Content-Type'] = 'application/json; charset=utf-8'
        if self.config.generate_client_request_id:
            header_parameters['x-ms-client-request-id'] = str(uuid.uuid1())
        if custom_headers:
            header_parameters.update(custom_headers)
        if self.config.accept_language is not None:
            header_parameters['accept-language'] = self._serialize.header("self.config.accept_language", self.config.accept_language, 'str')

        # Construct body
        body_content = self._serialize.body(blob_details, 'BlobDetails')

        # Construct and send request
        request = self._client.post(url, query_parameters, header_parameters, body_content)
        response = self._client.send(request, stream=False, **operation_config)

        if response.status_code not in [200]:
            raise models.RestErrorException(self._deserialize, response)

        deserialized = None

        if response.status_code == 200:
            deserialized = self._deserialize('SasUriResponse', response)

        if raw:
            client_raw_response = ClientRawResponse(deserialized, response)
            return client_raw_response

        return deserialized
    sas_uri.metadata = {'url': '/v1.0/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.Quantum/workspaces/{workspaceName}/storage/sasUri'}



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------

import uuid
from msrest.pipeline import ClientRawResponse
from msrestazure.azure_exceptions import CloudError

from .. import models


class JobsOperations(object):
    """JobsOperations operations.

    :param client: Client for service requests.
    :param config: Configuration of service client.
    :param serializer: An object model serializer.
    :param deserializer: An object model deserializer.
    """

    models = models

    def __init__(self, client, config, serializer, deserializer):

        self._client = client
        self._serialize = serializer
        self._deserialize = deserializer

        self.config = config

    def list(
            self, custom_headers=None, raw=False, **operation_config):
        """List jobs.

        :param dict custom_headers: headers that will be added to the request
        :param bool raw: returns the direct response alongside the
         deserialized response
        :param operation_config: :ref:`Operation configuration
         overrides<msrest:optionsforoperations>`.
        :return: An iterator like instance of JobDetails
        :rtype:
         ~azure.quantum.models.JobDetailsPaged[~azure.quantum.models.JobDetails]
        :raises: :class:`CloudError<msrestazure.azure_exceptions.CloudError>`
        """
        def internal_paging(next_link=None, raw=False):

            if not next_link:
                # Construct URL
                url = self.list.metadata['url']
                path_format_arguments = {
                    'subscriptionId': self._serialize.url("self.config.subscription_id", self.config.subscription_id, 'str'),
                    'resourceGroupName': self._serialize.url("self.config.resource_group_name", self.config.resource_group_name, 'str'),
                    'workspaceName': self._serialize.url("self.config.workspace_name", self.config.workspace_name, 'str')
                }
                url = self._client.format_url(url, **path_format_arguments)

                # Construct parameters
                query_parameters = {}

            else:
                url = next_link
                query_parameters = {}

            # Construct headers
            header_parameters = {}
            header_parameters['Accept'] = 'application/json'
            if self.config.generate_client_request_id:
                header_parameters['x-ms-client-request-id'] = str(uuid.uuid1())
            if custom_headers:
                header_parameters.update(custom_headers)
            if self.config.accept_language is not None:
                header_parameters['accept-language'] = self._serialize.header("self.config.accept_language", self.config.accept_language, 'str')

            # Construct and send request
            request = self._client.get(url, query_parameters, header_parameters)
            response = self._client.send(request, stream=False, **operation_config)

            if response.status_code not in [200]:
                exp = CloudError(response)
                exp.request_id = response.headers.get('x-ms-request-id')
                raise exp

            return response

        # Deserialize response
        deserialized = models.JobDetailsPaged(internal_paging, self._deserialize.dependencies)

        if raw:
            header_dict = {}
            client_raw_response = models.JobDetailsPaged(internal_paging, self._deserialize.dependencies, header_dict)
            return client_raw_response

        return deserialized
    list.metadata = {'url': '/v1.0/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.Quantum/workspaces/{workspaceName}/jobs'}

    def get(
            self, job_id, custom_headers=None, raw=False, **operation_config):
        """Get job by id.

        :param job_id: Id of the job.
        :type job_id: str
        :param dict custom_headers: headers that will be added to the request
        :param bool raw: returns the direct response alongside the
         deserialized response
        :param operation_config: :ref:`Operation configuration
         overrides<msrest:optionsforoperations>`.
        :return: JobDetails or ClientRawResponse if raw=true
        :rtype: ~azure.quantum.models.JobDetails or
         ~msrest.pipeline.ClientRawResponse
        :raises:
         :class:`RestErrorException<azure.quantum.models.RestErrorException>`
        """
        # Construct URL
        url = self.get.metadata['url']
        path_format_arguments = {
            'subscriptionId': self._serialize.url("self.config.subscription_id", self.config.subscription_id, 'str'),
            'resourceGroupName': self._serialize.url("self.config.resource_group_name", self.config.resource_group_name, 'str'),
            'workspaceName': self._serialize.url("self.config.workspace_name", self.config.workspace_name, 'str'),
            'jobId': self._serialize.url("job_id", job_id, 'str')
        }
        url = self._client.format_url(url, **path_format_arguments)

        # Construct parameters
        query_parameters = {}

        # Construct headers
        header_parameters = {}
        header_parameters['Accept'] = 'application/json'
        if self.config.generate_client_request_id:
            header_parameters['x-ms-client-request-id'] = str(uuid.uuid1())
        if custom_headers:
            header_parameters.update(custom_headers)
        if self.config.accept_language is not None:
            header_parameters['accept-language'] = self._serialize.header("self.config.accept_language", self.config.accept_language, 'str')

        # Construct and send request
        request = self._client.get(url, query_parameters, header_parameters)
        response = self._client.send(request, stream=False, **operation_config)

        if response.status_code not in [200]:
            raise models.RestErrorException(self._deserialize, response)

        deserialized = None

        if response.status_code == 200:
            deserialized = self._deserialize('JobDetails', response)

        if raw:
            client_raw_response = ClientRawResponse(deserialized, response)
            return client_raw_response

        return deserialized
    get.metadata = {'url': '/v1.0/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.Quantum/workspaces/{workspaceName}/jobs/{jobId}'}

    def create(
            self, job_id, job, custom_headers=None, raw=False, **operation_config):
        """Create a job.

        :param job_id: Id of the job.
        :type job_id: str
        :param job: The complete metadata of the job to submit.
        :type job: ~azure.quantum.models.JobDetails
        :param dict custom_headers: headers that will be added to the request
        :param bool raw: returns the direct response alongside the
         deserialized response
        :param operation_config: :ref:`Operation configuration
         overrides<msrest:optionsforoperations>`.
        :return: JobDetails or ClientRawResponse if raw=true
        :rtype: ~azure.quantum.models.JobDetails or
         ~msrest.pipeline.ClientRawResponse
        :raises:
         :class:`RestErrorException<azure.quantum.models.RestErrorException>`
        """
        # Construct URL
        url = self.create.metadata['url']
        path_format_arguments = {
            'subscriptionId': self._serialize.url("self.config.subscription_id", self.config.subscription_id, 'str'),
            'resourceGroupName': self._serialize.url("self.config.resource_group_name", self.config.resource_group_name, 'str'),
            'workspaceName': self._serialize.url("self.config.workspace_name", self.config.workspace_name, 'str'),
            'jobId': self._serialize.url("job_id", job_id, 'str')
        }
        url = self._client.format_url(url, **path_format_arguments)

        # Construct parameters
        query_parameters = {}

        # Construct headers
        header_parameters = {}
        header_parameters['Accept'] = 'application/json'
        header_parameters['Content-Type'] = 'application/json; charset=utf-8'
        if self.config.generate_client_request_id:
            header_parameters['x-ms-client-request-id'] = str(uuid.uuid1())
        if custom_headers:
            header_parameters.update(custom_headers)
        if self.config.accept_language is not None:
            header_parameters['accept-language'] = self._serialize.header("self.config.accept_language", self.config.accept_language, 'str')

        # Construct body
        body_content = self._serialize.body(job, 'JobDetails')

        # Construct and send request
        request = self._client.put(url, query_parameters, header_parameters, body_content)
        response = self._client.send(request, stream=False, **operation_config)

        if response.status_code not in [200, 201]:
            raise models.RestErrorException(self._deserialize, response)

        deserialized = None

        if response.status_code == 200:
            deserialized = self._deserialize('JobDetails', response)
        if response.status_code == 201:
            deserialized = self._deserialize('JobDetails', response)

        if raw:
            client_raw_response = ClientRawResponse(deserialized, response)
            return client_raw_response

        return deserialized
    create.metadata = {'url': '/v1.0/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.Quantum/workspaces/{workspaceName}/jobs/{jobId}'}

    def cancel(
            self, job_id, custom_headers=None, raw=False, **operation_config):
        """Cancel a job.

        :param job_id: Id of the job.
        :type job_id: str
        :param dict custom_headers: headers that will be added to the request
        :param bool raw: returns the direct response alongside the
         deserialized response
        :param operation_config: :ref:`Operation configuration
         overrides<msrest:optionsforoperations>`.
        :return: None or ClientRawResponse if raw=true
        :rtype: None or ~msrest.pipeline.ClientRawResponse
        :raises:
         :class:`RestErrorException<azure.quantum.models.RestErrorException>`
        """
        # Construct URL
        url = self.cancel.metadata['url']
        path_format_arguments = {
            'subscriptionId': self._serialize.url("self.config.subscription_id", self.config.subscription_id, 'str'),
            'resourceGroupName': self._serialize.url("self.config.resource_group_name", self.config.resource_group_name, 'str'),
            'workspaceName': self._serialize.url("self.config.workspace_name", self.config.workspace_name, 'str'),
            'jobId': self._serialize.url("job_id", job_id, 'str')
        }
        url = self._client.format_url(url, **path_format_arguments)

        # Construct parameters
        query_parameters = {}

        # Construct headers
        header_parameters = {}
        if self.config.generate_client_request_id:
            header_parameters['x-ms-client-request-id'] = str(uuid.uuid1())
        if custom_headers:
            header_parameters.update(custom_headers)
        if self.config.accept_language is not None:
            header_parameters['accept-language'] = self._serialize.header("self.config.accept_language", self.config.accept_language, 'str')

        # Construct and send request
        request = self._client.delete(url, query_parameters, header_parameters)
        response = self._client.send(request, stream=False, **operation_config)

        if response.status_code not in [204]:
            raise models.RestErrorException(self._deserialize, response)

        if raw:
            client_raw_response = ClientRawResponse(None, response)
            return client_raw_response
    cancel.metadata = {'url': '/v1.0/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.Quantum/workspaces/{workspaceName}/jobs/{jobId}'}



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------

from msrest.serialization import Model


class JobDetails(Model):
    """Job details.

    Variables are only populated by the server, and will be ignored when
    sending a request.

    All required parameters must be populated in order to send to Azure.

    :param id: The job id.
    :type id: str
    :param name: The job name. Is not required for the name to be unique and
     it's only used for display purposes.
    :type name: str
    :param container_uri: Required. The blob container SAS uri, the container
     is used to host job data.
    :type container_uri: str
    :param input_data_uri: The input blob SAS uri, if specified, it will
     override the default input blob in the container.
    :type input_data_uri: str
    :param input_data_format: Required. The format of the input data.
    :type input_data_format: str
    :param input_params: The input parameters for the job. JSON object used by
     the target solver. It is expected that the size of this object is small
     and only used to specify parameters for the execution target, not the
     input data.
    :type input_params: object
    :param provider_id: Required. The unique identifier for the provider.
    :type provider_id: str
    :param target: Required. The target identifier to run the job.
    :type target: str
    :param metadata: The job metadata. Metadata provides client the ability to
     store client-specific information
    :type metadata: dict[str, str]
    :param output_data_uri: The output blob SAS uri. When a job finishes
     successfully, results will be uploaded to this blob.
    :type output_data_uri: str
    :param output_data_format: The format of the output data.
    :type output_data_format: str
    :ivar status: The job status. Possible values include: 'Waiting',
     'Executing', 'Succeeded', 'Failed', 'Cancelled'
    :vartype status: str or ~azure.quantum.models.JobStatus
    :ivar creation_time: The creation time of the job.
    :vartype creation_time: datetime
    :ivar begin_execution_time: The time when the job began execution.
    :vartype begin_execution_time: datetime
    :ivar end_execution_time: The time when the job finished execution.
    :vartype end_execution_time: datetime
    :ivar cancellation_time: The time when a job was successfully cancelled.
    :vartype cancellation_time: datetime
    :ivar error_data: The error data for the job. This is expected only when
     Status 'Failed'.
    :vartype error_data: ~azure.quantum.models.ErrorData
    """

    _validation = {
        'container_uri': {'required': True},
        'input_data_format': {'required': True},
        'provider_id': {'required': True},
        'target': {'required': True},
        'status': {'readonly': True},
        'creation_time': {'readonly': True},
        'begin_execution_time': {'readonly': True},
        'end_execution_time': {'readonly': True},
        'cancellation_time': {'readonly': True},
        'error_data': {'readonly': True},
    }

    _attribute_map = {
        'id': {'key': 'id', 'type': 'str'},
        'name': {'key': 'name', 'type': 'str'},
        'container_uri': {'key': 'containerUri', 'type': 'str'},
        'input_data_uri': {'key': 'inputDataUri', 'type': 'str'},
        'input_data_format': {'key': 'inputDataFormat', 'type': 'str'},
        'input_params': {'key': 'inputParams', 'type': 'object'},
        'provider_id': {'key': 'providerId', 'type': 'str'},
        'target': {'key': 'target', 'type': 'str'},
        'metadata': {'key': 'metadata', 'type': '{str}'},
        'output_data_uri': {'key': 'outputDataUri', 'type': 'str'},
        'output_data_format': {'key': 'outputDataFormat', 'type': 'str'},
        'status': {'key': 'status', 'type': 'str'},
        'creation_time': {'key': 'creationTime', 'type': 'iso-8601'},
        'begin_execution_time': {'key': 'beginExecutionTime', 'type': 'iso-8601'},
        'end_execution_time': {'key': 'endExecutionTime', 'type': 'iso-8601'},
        'cancellation_time': {'key': 'cancellationTime', 'type': 'iso-8601'},
        'error_data': {'key': 'errorData', 'type': 'ErrorData'},
    }

    def __init__(self, **kwargs):
        super(JobDetails, self).__init__(**kwargs)
        self.id = kwargs.get('id', None)
        self.name = kwargs.get('name', None)
        self.container_uri = kwargs.get('container_uri', None)
        self.input_data_uri = kwargs.get('input_data_uri', None)
        self.input_data_format = kwargs.get('input_data_format', None)
        self.input_params = kwargs.get('input_params', None)
        self.provider_id = kwargs.get('provider_id', None)
        self.target = kwargs.get('target', None)
        self.metadata = kwargs.get('metadata', None)
        self.output_data_uri = kwargs.get('output_data_uri', None)
        self.output_data_format = kwargs.get('output_data_format', None)
        self.status = None
        self.creation_time = None
        self.begin_execution_time = None
        self.end_execution_time = None
        self.cancellation_time = None
        self.error_data = None



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
import json
import logging
import msal
import os
import re
import sys
import time
import requests

from azure.storage.blob import ContainerClient
from pathlib import Path
from typing import List, Optional

from msrest.authentication import Authentication, BasicTokenAuthentication
from azure.quantum._client import QuantumClient
from azure.quantum._client.operations import JobsOperations, StorageOperations
from azure.quantum._client.models import BlobDetails
from azure.quantum import Job
try:
    from .version import __version__
except:
    __version__ = "<unknown>"

logger = logging.getLogger(__name__)

__all__ = ['Workspace']

def sdk_environment(name):
    return 'AZURE_QUANTUM_ENV' in os.environ and os.environ['AZURE_QUANTUM_ENV'] == name

## Settings based on environment variables:
BASE_URL_FROM_ENV = os.environ['AZURE_QUANTUM_BASEURL'] if 'AZURE_QUANTUM_BASEURL' in os.environ else None
if  sdk_environment('dogfood'):
    logger.info("Using DOGFOOD configuration.")
    BASE_URL                = lambda location: BASE_URL_FROM_ENV or f"https://{location}.quantum-test.azure.com/"
    ARM_BASE_URL            = "https://api-dogfood.resources.windows-int.net/"
    AAD_CLIENT_ID           = "46a998aa-43d0-4281-9cbb-5709a507ac36" # Microsoft Quantum Development Kit
    AAD_SCOPES              = [ "api://dogfood.azure-quantum/Jobs.ReadWrite" ]

else:
    if sdk_environment('canary'):
        logger.info("Using CANARY configuration.")
        BASE_URL                = lambda location: BASE_URL_FROM_ENV or f"https://eastus2euap.quantum.azure.com/"
    else:
        logger.debug("Using production configuration.")
        BASE_URL                = lambda location: BASE_URL_FROM_ENV or f"https://{location}.quantum.azure.com/"
    ARM_BASE_URL            = "https://management.azure.com/"
    AAD_CLIENT_ID           = "84ba0947-6c53-4dd2-9ca9-b3694761521b" # Microsoft Quantum Development Kit
    AAD_SCOPES              = [ "https://quantum.microsoft.com/Jobs.ReadWrite" ]

TOKEN_CACHE = os.environ['AZURE_QUANTUM_TOKEN_CACHE'] if 'AZURE_QUANTUM_TOKEN_CACHE' in os.environ else os.path.join(Path.home(), ".azure-quantum", "aad.bin")

# Keeps track of the account name the current token is associated, so we only show
# a log message about acquiring a new token if the account changes.
_last_account = None

# Caches the MSAL Apps based on the subscription id
_msal_apps = {}

class TokenCacheWrapper:
    def __init__(self):
        self.cache_path  = TOKEN_CACHE
        self.initialize_cache()

    def initialize_cache(self):
        logger.debug(f"Using path '{self.cache_path}' for AAD token cache")

        self.token_cache = msal.SerializableTokenCache()
        if os.path.exists(self.cache_path):
            with open(self.cache_path, "r") as cache:
                self.token_cache.deserialize(cache.read())

    def write_out_cache(self):
        logger.debug(f"Updating AAD token cache at '{self.cache_path}'")

        cache_folder = os.path.dirname(self.cache_path)
        if not os.path.isdir(cache_folder):
            os.makedirs(cache_folder)

        with open(self.cache_path, "w") as afile:
            afile.write(self.token_cache.serialize())

class MsalWrapper:
    def __init__(self, subscription_id: str, refresh: bool):
        self.subscription_id     = subscription_id
        self.refresh             = refresh
        self.client_id           = AAD_CLIENT_ID
        self.scopes              = AAD_SCOPES
        self.token_cache_wrapper = TokenCacheWrapper()

    def get_tenant_authorization_uri(self):
        try:
            uri = f"{ARM_BASE_URL}/subscriptions/{self.subscription_id}?api-version=2018-01-01"
            response = requests.get(uri)

            # This gnarly piece of code is how we get the guest tenant authority associated with the subscription.
            # We make a unauthenticated request to ARM and extract the tenant authority from the WWW-Authenticate header in the response.
            # The header is of the form - Bearer authoritization_uri=https://login.microsoftonline.com/tenantId, key1=value1 
            auth_header = response.headers["WWW-Authenticate"]
            logger.debug (f"got the following auth header from the management endpoint: {auth_header}")

            trimmed_auth_header = auth_header[len("Bearer "):]                                  # trim the leading 'Bearer '
            trimmed_auth_header_parts = trimmed_auth_header.split(",")                            # get the various k=v parts
            key_value_pairs = dict(map(lambda s: tuple(s.split("=")), trimmed_auth_header_parts)) # make the parts into a dictionary
            quoted_tenant_uri = key_value_pairs["authorization_uri"]                             # get the value of the 'authorization_uri' key
            tenant_uri = quoted_tenant_uri[1:-1]                                                # strip it of surrounding quotes

            logger.debug (f"got the following tenant uri from the authentication header: {tenant_uri}")

            return tenant_uri

        except Exception as e:
            logger.debug(f"Failed to get tenant authority for subscription: {e}")
            return None

    def get_app(self):
        if not self.subscription_id in _msal_apps:
            authority = self.get_tenant_authorization_uri()
            if(authority == None):
                raise ValueError(f"Failed to get tenant authority for subscription '{self.subscription_id}'. Make sure the subscription id is correct.")

            try:
                _msal_apps[self.subscription_id] = msal.PublicClientApplication(
                    self.client_id,
                    authority   = authority,
                    token_cache = self.token_cache_wrapper.token_cache)
                logger.debug (f"Created a new app with the authority: {authority}")
            except Exception as e:
                 raise ValueError(f"Failed to create PublicClientApplication with tenant authority: {e}")
        return _msal_apps[self.subscription_id]

    def clear_accounts(self):
        accounts = self.get_app().get_accounts()
        for account in accounts:
            self.get_app().remove_account(account)

    def get_token_from_device_flow(self):
        # Clear accounts before doing device flow to make sure that we are left with a single account after the user completes the device flow.
        # We use that account is subsequent silent token acquisitions.
        self.clear_accounts()

        flow = self.get_app().initiate_device_flow(scopes=self.scopes)
        if "user_code" not in flow:
            raise ValueError("Fail to create device flow. Err: %s" % json.dumps(flow, indent=4))

        # Print a message for users to login via browser
        print(flow["message"])
        sys.stdout.flush()  # Some terminal needs this to ensure the message is shown

        result = self.get_app().acquire_token_by_device_flow(flow)  # Block until user has logged in and we're ready to continue:

        return result

    def try_get_token_silently(self):
        def try_get_account_from_app(app):
            accounts = app.get_accounts()
            if len(accounts) > 0:
                return accounts[0]
            else:
                return None

        account = try_get_account_from_app(self.get_app())

        if not account:
            return None

        # extract the tenant id from the home_account_id. The home_account_id is of the form "<oid>.<tenantId>"
        if "home_account_id" not in account:
            logger.debug("No home_account_id in account")
        
        tid = None
        account_parts = account["home_account_id"].split(".")
        if len(account_parts) < 2:
            logger.debug("No tenantId in account")
        else:
            tid = account_parts[1]

        global _last_account
        if _last_account != account['username']:
            _last_account = account['username']
            logger.info(f"Account(s) exists in aad token cache. Getting token for (userName: {account['username']}, tenantId: {tid})")

        return self.get_app().acquire_token_silent(self.scopes, account)

    def acquire_auth_token(self):
        """Returns an AAD token, either from a local cache or from AAD.

        It will first try to retrieve the token from a local cache
        (the location of the cache can be specified via the AZURE_QUANTUM_TOKEN_CACHE environent variable).
        If it the cache is empty or expired, then it uses the device token flow from MSAL (https://github.com/AzureAD/microsoft-authentication-library-for-python)
        to acquire the token from AAD.

        Once a token is received, it is inspected to see if it is a token for an MSA account.
        If so, the guest tenant id for that account is acquired, and a token is reaquired with that authority.

        If successful, it stores the acquired token in the cache for future calls.

        :param refresh:
            if true, by-passes the cache and fetches a fresh token from AAD.
        """
        result = None

        if self.refresh:
            logger.debug("Refresh forced. Skipping getting token from cache...")
            self.clear_accounts()
        else:
            logger.debug("Trying to get token from cache...")
            result = self.try_get_token_silently()

        if not result:
            logger.debug("...and trying to get a token from the device flow...")
            result = self.get_token_from_device_flow()
            logger.debug(f"...device flow returned: {result}")

        # At this point, there should be an access token; raise otherwise:
        if result and ("access_token" in result):
            logger.debug(f"Token:\n{result['access_token']}")

            # Store token back in cache:
            self.token_cache_wrapper.write_out_cache()
            return result
        else:
            raise ValueError("Failed to acquire AAD token.")

class Workspace:
    """Represents an Azure Quantum workspace.

    When creating a Workspace object, callers have two options for identifying
    the Azure Quantum workspace:
    1. specify a valid resource ID, or
    2. specify a valid subscription ID, resource group, and workspace name.
    
    If the Azure Quantum workspace does not have linked storage, the caller
    must also pass a valid Azure storage account connection string.

    :param subscription_id:
        The Azure subscription ID. Ignored if resource_id is specified.
        
    :param resource_group:
        The Azure resource group name. Ignored if resource_id is specified.
        
    :param name:
        The Azure Quantum workspace name. Ignored if resource_id is specified.
        
    :param storage:
        The Azure storage account connection string. Required only if the specified Azure Quantum
        workspace does not have linked storage.
        
    :param resource_id:
        The resource ID of the Azure Quantum workspace.
        
    :param location:
        The Azure region where the Azure Quantum workspace is provisioned.
        This may be specified as a region name such as \"East US\" or a location name such as \"eastus\".
        If no valid value is specified, defaults to \"westus\".
    """
    credentials = None

    def __init__(
        self,
        subscription_id: Optional[str] = None,
        resource_group: Optional[str] = None,
        name: Optional[str] = None,
        storage: Optional[str] = None,
        resource_id: Optional[str] = None,
        location: Optional[str] = None
    ):

        if resource_id is not None:
            # A valid resource ID looks like:
            # /subscriptions/f846b2bd-d0e2-4a1d-8141-4c6944a9d387/resourceGroups/RESOURCE_GROUP_NAME/providers/Microsoft.Quantum/Workspaces/WORKSPACE_NAME
            regex = r'^/subscriptions/([a-fA-F0-9-]*)/resourceGroups/([^\s/]*)/providers/Microsoft\.Quantum/Workspaces/([^\s/]*)$'
            match = re.search(regex, resource_id, re.IGNORECASE)
            if match:
                # match should contain four groups:
                # -> match.group(0): The full resource ID for the Azure Quantum workspace
                # -> match.group(1): The Azure subscription ID
                # -> match.group(2): The Azure resource group name
                # -> match.group(3): The Azure Quantum workspace name
                subscription_id = match.group(1)
                resource_group = match.group(2)
                name = match.group(3)

        if not subscription_id or not resource_group or not name:
            raise ValueError(
                "Azure Quantum workspace not fully specified. Please specify either a valid resource ID " +
                "or a valid combination of subscription ID, resource group name, and workspace name.")

        self.name = name
        self.resource_group = resource_group
        self.subscription_id = subscription_id
        self.storage = storage

        # Convert user-provided location into names recognized by Azure resource manager.
        # For example, a customer-provided value of "West US" should be converted to "westus".
        self.location = "".join(location.split()).lower() if location and location.split() else "westus"

    def _create_client(self) -> QuantumClient:
        auth = self.login()
        base_url = BASE_URL(self.location)
        logger.debug(f"Creating client for: subs:{self.subscription_id}, rg={self.resource_group}, ws={self.name}, frontdoor={base_url}")
        client = QuantumClient(auth, self.subscription_id, self.resource_group, self.name, base_url)
        return client

    def _create_jobs_client(self) -> JobsOperations:
        client = self._create_client().jobs
        return client

    def _create_workspace_storage_client(self) -> StorageOperations:
        client = self._create_client().storage
        return client

    def _custom_headers(self):
        return {
            'x-ms-azurequantum-sdk-version': __version__
        }

    def _get_linked_storage_sas_uri(self, container_name: str, blob_name: str=None) -> str:
        """
        Calls the service and returns a container sas url
        """
        client = self._create_workspace_storage_client()
        blob_details = BlobDetails(container_name=container_name, blob_name=blob_name)
        container_uri = client.sas_uri(blob_details=blob_details)

        logger.debug(f"Container URI from service: {container_uri}")
        return container_uri.sas_uri

    def submit_job(self, job: Job) -> Job:
        client = self._create_jobs_client()
        details = client.create(job.details.id, job.details, custom_headers=self._custom_headers())
        return Job(self, details)

    def cancel_job(self, job: Job) -> Job:
        client = self._create_jobs_client()
        client.cancel(job.details.id, custom_headers=self._custom_headers())
        details = client.get(job.id, custom_headers=self._custom_headers())
        return Job(self, details)

    def get_job(self, job_id: str) -> Job:
        """Returns the job corresponding to the given id."""
        client = self._create_jobs_client()
        details = client.get(job_id, custom_headers=self._custom_headers())
        return Job(self, details)

    def list_jobs(self) -> List[Job]:
        client = self._create_jobs_client()
        jobs = client.list(custom_headers=self._custom_headers())

        result = []
        for j in jobs:
            result.append(Job(self, j))

        return result

    def login(self, refresh:bool=False) -> Authentication:
        """Creates user credentials to authenticate with Azure Quantum

        It will first try to read the credentials from a secure local cache
        (the location of the cache can be specified via the AZURE_QUANTUM_TOKEN_CACHE environent variable).
        If it the cache is empty or expired, then it uses the device token flow from MSAL (https://github.com/AzureAD/microsoft-authentication-library-for-python)
        to acquire a token from AAD.

        If successful, it stores the acquired token in the cache for future calls.

        :param refresh:
            if true, by-passes the cache and fetches a fresh token from AAD.

        :returns:
            the user credentials to authenticate with Azure Quantum
        """

        # We only clear the credentials if they are of type BasicTokenAuthentication
        if refresh and isinstance(self.credentials, BasicTokenAuthentication):
            self.credentials = None

        if self.credentials is None:
            msal_wrapper = MsalWrapper(subscription_id=self.subscription_id, refresh=refresh)

            auth_token = msal_wrapper.acquire_auth_token()
            return BasicTokenAuthentication(token=auth_token)

        # When we already have an object of type Authentication, like
        # the ServicePrincipalCredentials, we should NOT wrap it in a BasicTokenAuthentication
        if not isinstance(self.credentials, Authentication):
            return BasicTokenAuthentication(token=self.credentials)


        return self.credentials



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##

import logging
import uuid
import io
import gzip
import json

from typing import List, Union, Dict, Optional, TYPE_CHECKING
from enum import Enum
from azure.quantum.optimization import Term
from azure.quantum.storage import upload_blob, ContainerClient


logger = logging.getLogger(__name__)

__all__ = ['Problem', 'ProblemType']

if TYPE_CHECKING:
    from azure.quantum.workspace import Workspace


class ProblemType(str, Enum):
    pubo  = 0
    ising = 1


class Problem:
    """Problem to submit to the service.

    :param name: Problem name
    :type name: str
    :param terms: Problem terms, depending on solver. Defaults to None
    :type terms: Optional[List[Term]], optional
    :param init_config: Optional configuration details, depending on solver. Defaults to None
    :type init_config: Optional[Dict[str,int]], optional
    :param problem_type: Problem type (ProblemType.pubo or ProblemType.ising), defaults to ProblemType.ising
    :type problem_type: ProblemType, optional
    """
    def __init__(
        self,
        name: str,
        terms: Optional[List[Term]] = None,
        init_config: Optional[Dict[str,int]] = None,
        problem_type: ProblemType = ProblemType.ising
    ):
        self.name = name
        self.terms = terms if terms is not None else []
        self.problem_type = problem_type
        self.init_config = init_config

    def serialize(self) -> str:
        result = {
            "cost_function": {
                "version": "1.1" if self.init_config else "1.0",
                "type": self.problem_type.name,
                "terms": [term.to_dict() for term in self.terms]
            }       
        }

        if self.init_config:
            result["cost_function"]["initial_configuration"] = self.init_config

        return json.dumps(result)

    def add_term(self, c: Union[int, float], indices: List[int]):
        self.terms.append(Term(indices=indices, c=c))
    
    def add_terms(self, terms: List[Term]):
        self.terms += terms

    def upload(
        self,
        workspace: "Workspace",
        container_name: str = "qio-problems",
        blob_name: str = None,
        compress: bool = True
    ):
        """Uploads an optimization problem instance to the cloud storage linked with the Workspace.

        :param workspace: interaction terms of the problem.
        :type workspace: Workspace
        :param container_name: [description], defaults to "qio-problems"
        :type container_name: str, optional
        :param blob_name: [description], defaults to None
        :type blob_name: str, optional
        :param compress: [description], defaults to True
        :type compress: bool, optional
        :return: uri of the uploaded problem
        :rtype: [type]
        """
        if blob_name is None:
            blob_name = '{}-{}'.format(self.name, uuid.uuid1())
            
        problem_json = self.serialize()
        logger.debug("Problem json: " + problem_json)
        
        content_type = "application/json"
        encoding = ""
        data = io.BytesIO()
        if compress:
            encoding = "gzip"
            with gzip.GzipFile(fileobj=data, mode='w') as fo:
                fo.write(problem_json.encode())
        else:
            data.write(problem_json.encode())

        if not workspace.storage:
            # No storage account is passed, use the linked one
            container_uri = workspace._get_linked_storage_sas_uri(container_name)
            container_client = ContainerClient.from_container_url(container_uri)
            return upload_blob(container_client, blob_name, content_type, encoding, data.getvalue(), return_sas_token=False)
        else:
            # Use the specified storage account
            container_client = ContainerClient.from_connection_string(workspace.storage, container_name)
            return upload_blob(container_client, blob_name, content_type, encoding, data.getvalue(), return_sas_token=True)



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------

from msrest.service_client import SDKClient
from msrest import Serializer, Deserializer
from msrestazure import AzureConfiguration
from .version import VERSION
from .operations.jobs_operations import JobsOperations
from .operations.providers_operations import ProvidersOperations
from .operations.storage_operations import StorageOperations
from .operations.quotas_operations import QuotasOperations
from . import models


class QuantumClientConfiguration(AzureConfiguration):
    """Configuration for QuantumClient
    Note that all parameters used to create this instance are saved as instance
    attributes.

    :param credentials: Credentials needed for the client to connect to Azure.
    :type credentials: :mod:`A msrestazure Credentials
     object<msrestazure.azure_active_directory>`
    :param subscription_id: The Azure subscription ID. This is a
     GUID-formatted string (e.g. 00000000-0000-0000-0000-000000000000)
    :type subscription_id: str
    :param resource_group_name: Name of an Azure resource group.
    :type resource_group_name: str
    :param workspace_name: Name of the workspace.
    :type workspace_name: str
    :param str base_url: Service URL
    """

    def __init__(
            self, credentials, subscription_id, resource_group_name, workspace_name, base_url=None):

        if credentials is None:
            raise ValueError("Parameter 'credentials' must not be None.")
        if subscription_id is None:
            raise ValueError("Parameter 'subscription_id' must not be None.")
        if resource_group_name is None:
            raise ValueError("Parameter 'resource_group_name' must not be None.")
        if workspace_name is None:
            raise ValueError("Parameter 'workspace_name' must not be None.")
        if not base_url:
            base_url = 'https://quantum.azure.com'

        super(QuantumClientConfiguration, self).__init__(base_url)

        self.add_user_agent('quantumclient/{}'.format(VERSION))
        self.add_user_agent('Azure-SDK-For-Python')

        self.credentials = credentials
        self.subscription_id = subscription_id
        self.resource_group_name = resource_group_name
        self.workspace_name = workspace_name


class QuantumClient(SDKClient):
    """Azure Quantum REST API client

    :ivar config: Configuration for client.
    :vartype config: QuantumClientConfiguration

    :ivar jobs: Jobs operations
    :vartype jobs: azure.quantum.operations.JobsOperations
    :ivar providers: Providers operations
    :vartype providers: azure.quantum.operations.ProvidersOperations
    :ivar storage: Storage operations
    :vartype storage: azure.quantum.operations.StorageOperations
    :ivar quotas: Quotas operations
    :vartype quotas: azure.quantum.operations.QuotasOperations

    :param credentials: Credentials needed for the client to connect to Azure.
    :type credentials: :mod:`A msrestazure Credentials
     object<msrestazure.azure_active_directory>`
    :param subscription_id: The Azure subscription ID. This is a
     GUID-formatted string (e.g. 00000000-0000-0000-0000-000000000000)
    :type subscription_id: str
    :param resource_group_name: Name of an Azure resource group.
    :type resource_group_name: str
    :param workspace_name: Name of the workspace.
    :type workspace_name: str
    :param str base_url: Service URL
    """

    def __init__(
            self, credentials, subscription_id, resource_group_name, workspace_name, base_url=None):

        self.config = QuantumClientConfiguration(credentials, subscription_id, resource_group_name, workspace_name, base_url)
        super(QuantumClient, self).__init__(self.config.credentials, self.config)

        client_models = {k: v for k, v in models.__dict__.items() if isinstance(v, type)}
        self.api_version = '2019-11-04-preview'
        self._serialize = Serializer(client_models)
        self._deserialize = Deserializer(client_models)

        self.jobs = JobsOperations(
            self._client, self.config, self._serialize, self._deserialize)
        self.providers = ProvidersOperations(
            self._client, self.config, self._serialize, self._deserialize)
        self.storage = StorageOperations(
            self._client, self.config, self._serialize, self._deserialize)
        self.quotas = QuotasOperations(
            self._client, self.config, self._serialize, self._deserialize)



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------

from enum import Enum


class JobStatus(str, Enum):

    waiting = "Waiting"
    executing = "Executing"
    succeeded = "Succeeded"
    failed = "Failed"
    cancelled = "Cancelled"


class ProviderAvailability(str, Enum):

    available = "Available"
    degraded = "Degraded"
    unavailable = "Unavailable"


class TargetAvailability(str, Enum):

    available = "Available"
    degraded = "Degraded"
    unavailable = "Unavailable"


class DimensionScope(str, Enum):

    workspace = "Workspace"
    subscription = "Subscription"


class MeterPeriod(str, Enum):

    none = "None"
    monthly = "Monthly"



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------

from msrest.serialization import Model


class SasUriResponse(Model):
    """Get SAS URL operation response.

    :param sas_uri: A URL with a SAS token to upload a blob for execution in
     the given workspace.
    :type sas_uri: str
    """

    _attribute_map = {
        'sas_uri': {'key': 'sasUri', 'type': 'str'},
    }

    def __init__(self, **kwargs):
        super(SasUriResponse, self).__init__(**kwargs)
        self.sas_uri = kwargs.get('sas_uri', None)



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
import logging

try:
    from .version import __version__
except:
    __version__ = "<unknown>"

logger = logging.getLogger(__name__)
logger.info(f"version: {__version__}")

from .job       import *
from .workspace import *



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------

from msrest.serialization import Model


class TargetStatus(Model):
    """Target status.

    Variables are only populated by the server, and will be ignored when
    sending a request.

    :ivar id: Target id.
    :vartype id: str
    :ivar current_availability: Target availability. Possible values include:
     'Available', 'Degraded', 'Unavailable'
    :vartype current_availability: str or
     ~azure.quantum.models.TargetAvailability
    :ivar average_queue_time: Average queue time in seconds.
    :vartype average_queue_time: long
    :ivar status_page: A page with detailed status of the provider.
    :vartype status_page: str
    """

    _validation = {
        'id': {'readonly': True},
        'current_availability': {'readonly': True},
        'average_queue_time': {'readonly': True},
        'status_page': {'readonly': True},
    }

    _attribute_map = {
        'id': {'key': 'id', 'type': 'str'},
        'current_availability': {'key': 'currentAvailability', 'type': 'str'},
        'average_queue_time': {'key': 'averageQueueTime', 'type': 'long'},
        'status_page': {'key': 'statusPage', 'type': 'str'},
    }

    def __init__(self, **kwargs) -> None:
        super(TargetStatus, self).__init__(**kwargs)
        self.id = None
        self.current_availability = None
        self.average_queue_time = None
        self.status_page = None



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------

from msrest.serialization import Model


class BlobDetails(Model):
    """Blob details.

    All required parameters must be populated in order to send to Azure.

    :param container_name: Required. The container name.
    :type container_name: str
    :param blob_name: The blob name.
    :type blob_name: str
    """

    _validation = {
        'container_name': {'required': True},
    }

    _attribute_map = {
        'container_name': {'key': 'containerName', 'type': 'str'},
        'blob_name': {'key': 'blobName', 'type': 'str'},
    }

    def __init__(self, *, container_name: str, blob_name: str=None, **kwargs) -> None:
        super(BlobDetails, self).__init__(**kwargs)
        self.container_name = container_name
        self.blob_name = blob_name



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------

try:
    from .error_data_py3 import ErrorData
    from .job_details_py3 import JobDetails
    from .blob_details_py3 import BlobDetails
    from .sas_uri_response_py3 import SasUriResponse
    from .target_status_py3 import TargetStatus
    from .provider_status_py3 import ProviderStatus
    from .quota_py3 import Quota
    from .rest_error_py3 import RestError, RestErrorException
except (SyntaxError, ImportError):
    from .error_data import ErrorData
    from .job_details import JobDetails
    from .blob_details import BlobDetails
    from .sas_uri_response import SasUriResponse
    from .target_status import TargetStatus
    from .provider_status import ProviderStatus
    from .quota import Quota
    from .rest_error import RestError, RestErrorException
from .job_details_paged import JobDetailsPaged
from .provider_status_paged import ProviderStatusPaged
from .quota_paged import QuotaPaged
from .quantum_client_enums import (
    JobStatus,
    ProviderAvailability,
    TargetAvailability,
    DimensionScope,
    MeterPeriod,
)

__all__ = [
    'ErrorData',
    'JobDetails',
    'BlobDetails',
    'SasUriResponse',
    'TargetStatus',
    'ProviderStatus',
    'Quota',
    'RestError', 'RestErrorException',
    'JobDetailsPaged',
    'ProviderStatusPaged',
    'QuotaPaged',
    'JobStatus',
    'ProviderAvailability',
    'TargetAvailability',
    'DimensionScope',
    'MeterPeriod',
]



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------

from msrest.serialization import Model


class ErrorData(Model):
    """An error response from Azure.

    :param code: An identifier for the error. Codes are invariant and are
     intended to be consumed programmatically.
    :type code: str
    :param message: A message describing the error, intended to be suitable
     for displaying in a user interface.
    :type message: str
    """

    _attribute_map = {
        'code': {'key': 'code', 'type': 'str'},
        'message': {'key': 'message', 'type': 'str'},
    }

    def __init__(self, **kwargs):
        super(ErrorData, self).__init__(**kwargs)
        self.code = kwargs.get('code', None)
        self.message = kwargs.get('message', None)



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------

import uuid
from msrest.pipeline import ClientRawResponse

from .. import models


class ProvidersOperations(object):
    """ProvidersOperations operations.

    :param client: Client for service requests.
    :param config: Configuration of service client.
    :param serializer: An object model serializer.
    :param deserializer: An object model deserializer.
    """

    models = models

    def __init__(self, client, config, serializer, deserializer):

        self._client = client
        self._serialize = serializer
        self._deserialize = deserializer

        self.config = config

    def get_status(
            self, custom_headers=None, raw=False, **operation_config):
        """Get provider status.

        :param dict custom_headers: headers that will be added to the request
        :param bool raw: returns the direct response alongside the
         deserialized response
        :param operation_config: :ref:`Operation configuration
         overrides<msrest:optionsforoperations>`.
        :return: An iterator like instance of ProviderStatus
        :rtype:
         ~azure.quantum.models.ProviderStatusPaged[~azure.quantum.models.ProviderStatus]
        :raises:
         :class:`RestErrorException<azure.quantum.models.RestErrorException>`
        """
        def internal_paging(next_link=None, raw=False):

            if not next_link:
                # Construct URL
                url = self.get_status.metadata['url']
                path_format_arguments = {
                    'subscriptionId': self._serialize.url("self.config.subscription_id", self.config.subscription_id, 'str'),
                    'resourceGroupName': self._serialize.url("self.config.resource_group_name", self.config.resource_group_name, 'str'),
                    'workspaceName': self._serialize.url("self.config.workspace_name", self.config.workspace_name, 'str')
                }
                url = self._client.format_url(url, **path_format_arguments)

                # Construct parameters
                query_parameters = {}

            else:
                url = next_link
                query_parameters = {}

            # Construct headers
            header_parameters = {}
            header_parameters['Accept'] = 'application/json'
            if self.config.generate_client_request_id:
                header_parameters['x-ms-client-request-id'] = str(uuid.uuid1())
            if custom_headers:
                header_parameters.update(custom_headers)
            if self.config.accept_language is not None:
                header_parameters['accept-language'] = self._serialize.header("self.config.accept_language", self.config.accept_language, 'str')

            # Construct and send request
            request = self._client.get(url, query_parameters, header_parameters)
            response = self._client.send(request, stream=False, **operation_config)

            if response.status_code not in [200]:
                raise models.RestErrorException(self._deserialize, response)

            return response

        # Deserialize response
        deserialized = models.ProviderStatusPaged(internal_paging, self._deserialize.dependencies)

        if raw:
            header_dict = {}
            client_raw_response = models.ProviderStatusPaged(internal_paging, self._deserialize.dependencies, header_dict)
            return client_raw_response

        return deserialized
    get_status.metadata = {'url': '/v1.0/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.Quantum/workspaces/{workspaceName}/providerStatus'}



##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
import logging
import io
import azure.quantum

from typing import List, Union, Any, Optional
from enum import Enum
from azure.quantum import Workspace, Job
from azure.quantum._client.models import JobDetails
from azure.quantum.optimization import Problem
from azure.quantum.storage import ContainerClient, create_container_using_client

logger = logging.getLogger(__name__)

__all__ = [
    'Solver',
    'ParallelTempering',
    'SimulatedAnnealing',
    'HardwarePlatform',
    'Tabu',
    'QuantumMonteCarlo'
]

class Solver:

    def __init__(
        self,
        workspace: Workspace,
        provider: str,
        target: str,
        input_data_format: str,
        output_data_format: str,
        nested_params: bool=True,
        force_str_params: bool=False
        ):
        self.workspace = workspace
        self.target = target
        self.provider = provider
        self.input_data_format = input_data_format
        self.output_data_format = output_data_format
        self.nested_params = nested_params
        self.force_str_params = force_str_params
        self.params = { "params": {} } if nested_params else {}

    def submit(self, problem: Union[str, Problem]) -> Job:
        """Submits a job to execution to the associated Azure Quantum Workspace.

       :param problem: 
            The Problem to solve. It can be an instance of a Problem, 
            or the URL of an Azure Storage Blob where the serialized version
            of a Problem has been uploaded.
        """
        ## Create a container URL:
        job_id = Job.create_job_id()
        logger.info(f"Submitting job with id: {job_id}")

        container_name = f"job-{job_id}"

        if not self.workspace.storage:
            # No storage account is passed, in this case, get linked account from the service
            container_uri = self.workspace._get_linked_storage_sas_uri(container_name)
            container_client = ContainerClient.from_container_url(container_uri)
            create_container_using_client(container_client)
            container_uri = azure.quantum.storage.remove_sas_token(container_uri)
        else:
            # Storage account is passed, use it to generate a container_uri
            container_uri = azure.quantum.storage.get_container_uri(self.workspace.storage, container_name)
        
        logger.debug(f"Container URI: {container_uri}")

        if isinstance(problem, str):
            name = "Optimization problem"
            problem_uri = problem  
        else:
            name = problem.name
            problem_uri = problem.upload(self.workspace, compress=True, container_name=container_name, blob_name="inputData")
        
        logger.info(f"Submitting problem '{name}'. Using payload from: '{problem_uri}'")

        details = JobDetails(
            id=job_id,
            name=name,
            container_uri=container_uri,
            input_data_format=self.input_data_format,
            output_data_format=self.output_data_format,
            input_data_uri=problem_uri,
            provider_id=self.provider,
            target=self.target,
            input_params=self.params
        )

        logger.debug(f"==> submitting: {details}")
        job = self.workspace.submit_job(Job(self.workspace, details))
        return job

    def optimize(self, problem: Union[str, Problem]):
        """Submits the Problem to the associated Azure Quantum Workspace and get the results. 
        
        :param problem: 
            The Problem to solve. It can be an instance of a Problem, 
            or the URL of an Azure Storage Blob where the serialized version
            of a Problem has been uploaded.
        """
        job = self.submit(problem)
        logger.info(f"Submitted job: '{job.id}'")
        
        return job.get_results()

    def set_one_param(self, name: str, value: Any):
        if value is not None:
            params = self.params["params"] if self.nested_params else self.params
            params[name] = str(value) if self.force_str_params else value

class HardwarePlatform(Enum):
    CPU = 1
    FPGA = 2

class ParallelTempering(Solver):
    def __init__(
        self,
        workspace: Workspace,
        sweeps: Optional[int] = None,
        replicas: Optional[int] = None,
        all_betas: Optional[List[int]] = None,
        timeout: Optional[int] = None,
        seed: Optional[int] = None):
        """The constructor of a Parallel Tempering solver.

        Multi-core Parallel Tempering solver for binary optimization problems with k-local interactions on an all-to-all 
        graph topology with double precision support for the coupler weights.

        :param sweeps: 
            specifies the number of sweeps.
        :param replicas: 
            specifies the number of replicas.
        :param all_betas: 
            a list of floats specifying the list of inverse temperatures. 
            > Note: this list must be equal in length to the number of replicas.            
        :param timeout: 
            specifies maximum number of seconds to run the core solver
            loop. initialization time does not respect this value, so the 
            solver may run longer than the value specified.
        :param seed: 
            specifies a random seed value.
        """
        param_free = (
            sweeps is None and
            replicas is None and
            all_betas is None
        )
        platform = HardwarePlatform.CPU
        if platform == HardwarePlatform.FPGA:
            target = "microsoft.paralleltempering.fpga" 
        else:
            target = "microsoft.paralleltempering-parameterfree.cpu" if param_free  else "microsoft.paralleltempering.cpu"

        super().__init__(
            workspace=workspace,
            provider="Microsoft",
            target=target,
            input_data_format="microsoft.qio.v2",
            output_data_format="microsoft.qio-results.v2")

        self.set_one_param("sweeps", sweeps)
        self.set_one_param("replicas", replicas)
        self.set_one_param("all_betas", all_betas)
        self.set_one_param("seed", seed)
        self.set_one_param("timeout", timeout)
        self.set_one_param("seed", seed)

        # Check parameters:
        if not all_betas is None:
            if replicas is None:
                self.set_one_param("replicas", len(all_betas))
            elif len(all_betas) != replicas:
                raise ValueError("Parameter 'replicas' must equal the length of the 'all_betas' parameters.")


class SimulatedAnnealing(Solver):
    
    def __init__(
        self,
        workspace: Workspace,
        beta_start: Optional[float]=None,
        beta_stop: Optional[float]=None,
        sweeps: Optional[int]=None,
        restarts: Optional[int]=None,
        timeout: Optional[int]=None,
        seed: Optional[int]=None,
        platform: Optional[HardwarePlatform]=HardwarePlatform.CPU
        ):
        """The constructor of an Simulated Annealing solver.
        
        Multi-core Simulated Annealing solver for binary optimization problems 
        with k-local interactions on an all-to-all graph topology with double 
        precision support for the coupler weights. 

        :param beta_start: 
            specifies the starting inverse temperature.
        :param beta_stop: 
            specifies the stopping inverse temperature.
        :param sweeps: 
            specifies the number of sweeps.
        :param restarts: 
            specifies the number of restarts.
        :param timeout: 
            specifies maximum number of seconds to run the core solver
            loop. initialization time does not respect this value, so the 
            solver may run longer than the value specified.
        :param seed: 
            specifies a random seed value.
        :platform:
            specifies hardware platform HardwarePlatform.CPU or HardwarePlatform.FPGA.
        """
        param_free = (
            beta_start is None and
            beta_stop is None and 
            sweeps is None and
            restarts is None 
        )

        if platform == HardwarePlatform.FPGA:
            target = "microsoft.simulatedannealing-parameterfree.fpga" if param_free  else "microsoft.simulatedannealing.fpga"
        else:
            target = "microsoft.simulatedannealing-parameterfree.cpu" if param_free  else "microsoft.simulatedannealing.cpu"
        
        super().__init__(
            workspace=workspace,
            provider="Microsoft",
            target=target,
            input_data_format="microsoft.qio.v2",
            output_data_format="microsoft.qio-results.v2")

        self.set_one_param("beta_start", beta_start)
        self.set_one_param("beta_stop", beta_stop)
        self.set_one_param("sweeps", sweeps)
        self.set_one_param("restarts", restarts)
        self.set_one_param("timeout", timeout)
        self.set_one_param("seed", seed)


class Tabu(Solver):
    
    def __init__(
        self,
        workspace: Workspace,
        sweeps: Optional[int]=None,
        tabu_tenure: Optional[int]=None,
        timeout: Optional[int]=None,
        seed: Optional[int]=None
    ):
        """The constructor of an Tabu Search solver.
        
        Multi-core Tabu Search solver for binary optimization problems 
        with k-local interactions on an all-to-all graph topology with double 
        precision support for the coupler weights. 

        This solver is CPU only.

        :param sweeps: 
            specifies the number of sweeps.
        :param tabu_tenure: 
            specifies the tabu tenure.
        :param timeout: 
            specifies maximum number of seconds to run the core solver
            loop. initialization time does not respect this value, so the 
            solver may run longer than the value specified.
        :param seed: 
            specifies a random seed value.
        """
        param_free = (
            sweeps is None and
            tabu_tenure is None 
        )
        
        target = "microsoft.tabu-parameterfree.cpu" if param_free else "microsoft.tabu.cpu"
        
        super().__init__(
            workspace=workspace, 
            provider="Microsoft", 
            target=target,
            input_data_format="microsoft.qio.v2",
            output_data_format="microsoft.qio-results.v2")
        
        self.set_one_param("sweeps", sweeps)
        self.set_one_param("tabu_tenure", tabu_tenure)
        self.set_one_param("timeout", timeout)
        self.set_one_param("seed", seed)

class QuantumMonteCarlo(Solver):
    def __init__(
        self, 
        workspace: Workspace, 
        trotter_number: Optional[int]=None, 
        seed: Optional[int]=None,  
        transverse_field_start: Optional[float] = None,
        transverse_field_stop: Optional[float] = None,
        restarts: Optional[int] = None,
        sweeps: Optional[int]=None,
        beta_start: Optional[float]=None
    ):
        """The constructor of Simulated Qunatum Annelaing Search solver.
        
        Simulated Quantum Search solver for binary optimization problems 
        with k-local interactions on an all-to-all graph topology with double 
        precision support for the coupler weights. 

        This solver is CPU only.

        :param trotter_number: 
            specifies the number of trotter time slices i.e. number of copies of each variable.
        :param seed: 
            specifies a random seed value.
        :param sweeps:
            Number of monte carlo sweeps
        :param transverse_field_start:
            starting strength of the external field
        :param transverse_field_end:
            ending strength of the external field
        :param beta_start: 
            Low starting temp i.e. beta start 
        :param restarts:
            Number of simulation runs
        """
        
        target = "microsoft.qmc.cpu"
        super().__init__(
            workspace=workspace,
            provider="Microsoft",
            target=target,
            input_data_format="microsoft.qio.v2",
            output_data_format="microsoft.qio-results.v2")
        self.set_one_param("sweeps", sweeps)
        self.set_one_param("trotter_number", trotter_number)
        self.set_one_param("seed", seed)
        self.set_one_param("transverse_field_start",transverse_field_start)
        self.set_one_param("transverse_field_stop",transverse_field_stop)
        self.set_one_param("beta_start",beta_start)
        self.set_one_param("restarts",restarts)




##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
import io
import logging
from typing import Any

from azure.storage.blob import BlobServiceClient, ContainerClient, BlobClient, BlobSasPermissions, ContentSettings, generate_blob_sas, generate_container_sas
from datetime import datetime, timedelta

logger = logging.getLogger(__name__)

def create_container(connection_string: str, container_name: str) -> ContainerClient:
    """
    Creates and initialize a container; returns the client needed to access it.
    """
    blob_service_client = BlobServiceClient.from_connection_string(connection_string)
    logger.info(f"Initializing storage client for account: '{blob_service_client.account_name}'")

    container_client = blob_service_client.get_container_client(container_name)
    create_container_using_client(container_client)
    return container_client

def create_container_using_client(container_client: ContainerClient):
    """
    Creates and initializes a container.
    """
    try:
        container_client.get_container_properties()
        logger.debug(f"  - uploading to existing container")
    except:
        logger.debug(f"  - uploading to **new** container: {container_client.container_name}")
        container_client.create_container()

def get_container_uri(connection_string: str, container_name: str) -> str:
    """
    Creates and initialize a container; returns a URI with a SAS read/write token to access it.
    """
    container = create_container(connection_string, container_name)
    logger.info(f"Creating SAS token for container '{container_name}' on account: '{container.account_name}'")

    sas_token = generate_container_sas(
        container.account_name,
        container.container_name,
        account_key=container.credential.account_key,
        permission=BlobSasPermissions(read=True, add=True, write=True, create=True),
        expiry= datetime.utcnow() + timedelta(days=14)
    )

    uri = container.url + "?" + sas_token
    logger.debug(f"  - container url: '{uri}'.")
    return uri


def upload_blob(
    container: ContainerClient,
    blob_name: str,
    content_type:str,
    content_encoding:str,
    data: Any,
    return_sas_token: bool=True
) -> str:
    """
    Uploads the given data to a blob record. If a blob with the given name already exist, it throws an error.

    Returns a uri with a SAS token to access the newly created blob.
    """
    create_container_using_client(container)
    logger.info(f"Uploading blob '{blob_name}' to container '{container.container_name}' on account: '{container.account_name}'")

    content_settings = ContentSettings(content_type=content_type, content_encoding=content_encoding)
    blob = container.get_blob_client(blob_name)
    blob.upload_blob(data, content_settings=content_settings)
    logger.debug(f"  - blob '{blob_name}' uploaded. generating sas token.")

    if return_sas_token:
        sas_token = generate_blob_sas(
            blob.account_name,
            blob.container_name,
            blob.blob_name,
            account_key=blob.credential.account_key,
            permission=BlobSasPermissions(read=True),
            expiry=datetime.utcnow() + timedelta(days=14)
        )

        uri = blob.url + "?" + sas_token
    else:
        uri = remove_sas_token(blob.url)

    logger.debug(f"  - blob access url: '{uri}'.")

    return uri

def download_blob(blob_url: str) -> Any:
    """
    Downloads the given blob from the container.
    """
    blob_client = BlobClient.from_blob_url(blob_url)
    logger.info(f"Downloading blob '{blob_client.blob_name}' from container '{blob_client.container_name}' on account: '{blob_client.account_name}'")

    response = blob_client.download_blob().readall()
    logger.debug(response)

    return response

def remove_sas_token(sas_uri: str) -> str:
    index = sas_uri.find('?')
    if index != -1:
        sas_uri = sas_uri[0:index]
    
    return sas_uri




##
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
##
import logging
import time
import io
import json
import uuid

from typing import List, TYPE_CHECKING
from urllib.parse import urlparse

from msrest.authentication import Authentication
from azure.quantum._client.models import JobDetails
from azure.quantum.storage import download_blob
from azure.storage.blob import BlobServiceClient, ContainerClient, BlobClient, BlobSasPermissions, generate_blob_sas, generate_container_sas

__all__ = ['Job']

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from azure.quantum.workspace import Workspace

class Job:
    """Azure Quantum Job that is submitted to a given Workspace.

    :param workspace: Workspace instance to submit job to
    :type workspace: Workspace
    :param job_details: Job details model, contains Job ID, name and other details
    :type job_details: JobDetails
    """
    def __init__(self, workspace: "Workspace", job_details: JobDetails):
      self.workspace = workspace
      self.details = job_details
      self.id = job_details.id
      self.results = None

    def refresh(self):
        """Refreshes the Job's details by querying the workspace.
        """
        self.details = self.workspace.get_job(self.id).details

    def has_completed(self):
        return (
            self.details.status == "Succeeded" or
            self.details.status == "Failed" or
            self.details.status == "Cancelled"
        )

    def wait_until_completed(self, max_poll_wait_secs=30):
        """Keeps refreshing the Job's details until it reaches a finished status.
        """
        self.refresh()
        poll_wait = 0.2
        while not self.has_completed():
            logger.debug(f"Waiting for job {self.id}, it is in status '{self.details.status}'")
            print('.', end='', flush=True)
            time.sleep(poll_wait)
            self.refresh()
            poll_wait = max_poll_wait_secs if poll_wait >= max_poll_wait_secs else poll_wait * 1.5

    def get_results(self):
        if not self.results is None:
            return self.results

        if not self.has_completed():
            self.wait_until_completed()

        if not self.details.status == "Succeeded":
            raise RuntimeError(f"Cannot retrieve results as job execution failed (status: {self.details.status}. error: {self.details.error_data})")

        url = urlparse(self.details.output_data_uri) 
        if url.query.find('se=') == -1:
            # output_data_uri does not contains SAS token, get sas url from service
            blob_client = BlobClient.from_blob_url(self.details.output_data_uri)
            blob_uri = self.workspace._get_linked_storage_sas_uri(blob_client.container_name, blob_client.blob_name)
            payload = download_blob(blob_uri)
        else:
            # output_data_uri contains SAS token, use it
            payload = download_blob(self.details.output_data_uri)

        result = json.loads(payload.decode('utf8'))
        return result
            
    @staticmethod
    def create_job_id() -> str:
        """Create a unique id for a new job.
        """
        return str(uuid.uuid1())



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------

from msrest.serialization import Model
from msrest.exceptions import HttpOperationError


class RestError(Model):
    """Error information returned by the API.

    :param error:
    :type error: ~azure.quantum.models.ErrorData
    """

    _attribute_map = {
        'error': {'key': 'error', 'type': 'ErrorData'},
    }

    def __init__(self, *, error=None, **kwargs) -> None:
        super(RestError, self).__init__(**kwargs)
        self.error = error


class RestErrorException(HttpOperationError):
    """Server responsed with exception of type: 'RestError'.

    :param deserialize: A deserializer
    :param response: Server response to be deserialized.
    """

    def __init__(self, deserialize, response, *args):

        super(RestErrorException, self).__init__(deserialize, response, 'RestError', *args)



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------

from msrest.serialization import Model


class ProviderStatus(Model):
    """Providers status.

    Variables are only populated by the server, and will be ignored when
    sending a request.

    :ivar id: Provider id.
    :vartype id: str
    :ivar current_availability: Provider availability. Possible values
     include: 'Available', 'Degraded', 'Unavailable'
    :vartype current_availability: str or
     ~azure.quantum.models.ProviderAvailability
    :ivar targets:
    :vartype targets: list[~azure.quantum.models.TargetStatus]
    """

    _validation = {
        'id': {'readonly': True},
        'current_availability': {'readonly': True},
        'targets': {'readonly': True},
    }

    _attribute_map = {
        'id': {'key': 'id', 'type': 'str'},
        'current_availability': {'key': 'currentAvailability', 'type': 'str'},
        'targets': {'key': 'targets', 'type': '[TargetStatus]'},
    }

    def __init__(self, **kwargs) -> None:
        super(ProviderStatus, self).__init__(**kwargs)
        self.id = None
        self.current_availability = None
        self.targets = None
